#include <windows.h>
#include <math.h>
#include <stdint.h>
#include <stdio.h>

#define CONFIGFILE "./GrimCam.ini"

#define DEG_TO_RAD (M_PI / 180.0)
#define TWO_PI (2.0 * M_PI)

namespace GAME
{
	class GameCamera
	{
	public:
		void *vtable;
		float distance;
		float yaw;
		float pitch;
		float fov;
		float far_plane;
		float near_plane;

		DECLSPEC_IMPORT void __thiscall SetMovementExtents(float, float, float, float, float);
	};

	class GameEngine
	{
	public:
		DECLSPEC_IMPORT GameCamera * __thiscall GetCamera(void);
	};

	class GraphicsEngine
	{
	public:
		DECLSPEC_IMPORT void __thiscall EnableFog(bool);
	};

	class Engine
	{
	public:
		DECLSPEC_IMPORT GraphicsEngine * __thiscall GetGraphicsEngine(void);
	};

	extern DECLSPEC_IMPORT GameEngine *gGameEngine;
	extern DECLSPEC_IMPORT Engine *gEngine;
}

static const char ret4[3] = { 0xC2, 0x04, 0x00 };

static int key = VK_MBUTTON; 
static float stepx = M_PI / 900.0;
static float stepy = M_PI / 900.0;
static float min_pitch = -89.0 * DEG_TO_RAD;
static float max_pitch = 89.0 * DEG_TO_RAD;

extern "C" BOOL WINAPI DllMain(HINSTANCE hinst, DWORD reason, LPVOID lptr)
{
	static void *set_yaw = NULL;
	static void *set_pitch = NULL;
	static void *set_far = NULL;
	static char saved_yaw[3];
	static char saved_pitch[3];
	static char saved_far[3];
	char buffer[32];
	HANDLE proc = GetCurrentProcess();
	HMODULE dll;

	switch(reason)
	{
	case DLL_PROCESS_ATTACH:
		dll = GetModuleHandle("engine.dll");
		if (dll)
		{
			set_yaw = (void *)GetProcAddress(dll, "?SetCameraYaw@WorldCamera@GAME@@UAEXM@Z");
			if (set_yaw)
			{
				ReadProcessMemory(proc, set_yaw, saved_yaw, 3, NULL);
				WriteProcessMemory(proc, set_yaw, ret4, 3, NULL);
			}

			set_pitch = (void *)GetProcAddress(dll, "?SetCameraPitch@WorldCamera@GAME@@QAEXM@Z");
			if (set_pitch)
			{
				ReadProcessMemory(proc, set_pitch, saved_pitch, 3, NULL);
				WriteProcessMemory(proc, set_pitch, ret4, 3, NULL);
			}

			set_far = (void *)GetProcAddress(dll, "?SetCameraFarPlane@WorldCamera@GAME@@QAEXM@Z");
			if (set_far)
			{
				ReadProcessMemory(proc, set_far, saved_far, 3, NULL);
				WriteProcessMemory(proc, set_far, ret4, 3, NULL);
			}
		}

		if (GAME::gGameEngine)
		{
			GAME::GameCamera *camera = GAME::gGameEngine->GetCamera();

			GetPrivateProfileString("Settings", "MinPitch", "-89", buffer, 32, CONFIGFILE);
			min_pitch = atof(buffer) * DEG_TO_RAD;
			GetPrivateProfileString("Settings", "MaxPitch", "89", buffer, 32, CONFIGFILE);
			max_pitch = atof(buffer) * DEG_TO_RAD;
			GetPrivateProfileString("Settings", "FOV", "60", buffer, 32, CONFIGFILE);
			camera->fov = atof(buffer) * (M_PI / 360.0);
			GetPrivateProfileString("Settings", "FarClip", "120", buffer, 32, CONFIGFILE);
			camera->far_plane = atof(buffer);
			GetPrivateProfileString("Settings", "HSensitivity", "0.2", buffer, 32, CONFIGFILE);
			stepx = atof(buffer) * DEG_TO_RAD;
			GetPrivateProfileString("Settings", "VSensitivity", "0.2", buffer, 32, CONFIGFILE);
			stepy = atof(buffer) * DEG_TO_RAD;
			key = GetPrivateProfileInt("Settings", "Key", VK_MBUTTON, CONFIGFILE);

			float min, max, fog;

			GetPrivateProfileString("Settings", "MinDistance", "20", buffer, 32, CONFIGFILE);
			min = atof(buffer);
			GetPrivateProfileString("Settings", "MaxDistance", "40", buffer, 32, CONFIGFILE);
			max = atof(buffer);
			GetPrivateProfileString("Settings", "DepthFogClamp", "0", buffer, 32, CONFIGFILE);
			fog = atof(buffer);

			camera->SetMovementExtents(min, max, 30.0, 50.0, fog);
		}

		if (GAME::gEngine)
		{
			int f = GetPrivateProfileInt("Settings", "EnableFog", 1, CONFIGFILE);
			GAME::gEngine->GetGraphicsEngine()->EnableFog(f);
		}


		break;

	case DLL_PROCESS_DETACH:
		if (GetModuleHandle("engine.dll"))
		{
			if (set_yaw)
			{
				WriteProcessMemory(proc, set_yaw, saved_yaw, 3, NULL);
			}

			if (set_pitch)
			{
				WriteProcessMemory(proc, set_pitch, saved_pitch, 3, NULL);
			}

			if (set_far)
			{
				WriteProcessMemory(proc, set_far, saved_far, 3, NULL);
			}
		}

		break;
	}

	return TRUE;
}

extern "C" DECLSPEC_EXPORT LRESULT CALLBACK Grim_MouseHook(int code, WPARAM wp, LPARAM lp)
{
	static BOOL btn = 0;
	static LONG x, y;
	PMOUSEHOOKSTRUCT m = (PMOUSEHOOKSTRUCT)lp;

	if (code == HC_ACTION && GAME::gGameEngine)
	{
		GAME::GameCamera *camera = GAME::gGameEngine->GetCamera();

		if (GetKeyState(key) < 0)
		{
			if (btn)
			{
				LONG dx = m->pt.x - x;
				LONG dy = m->pt.y - y;

				if (dx)
				{
					float yaw = camera->yaw - dx * stepx;
					while (yaw < 0.0) yaw += TWO_PI;
					while (yaw > TWO_PI) yaw -= TWO_PI;
					camera->yaw = yaw;
				}

				if (dy)
				{
					float pitch = camera->pitch + dy * stepy;
					if (pitch < min_pitch) pitch = min_pitch;
					if (pitch > max_pitch) pitch = max_pitch;
					camera->pitch = pitch;
				}
			}

			btn = 1;
			x = m->pt.x;
			y = m->pt.y;
		}
		else
		{
			btn = 0;
		}
	}

	return CallNextHookEx(0, code, wp, lp);
}

